/**
 * @file GMultiConstraintT.hpp
 */

/*
 * Copyright (C) Gemfony scientific UG (haftungsbeschraenkt)
 *
 * See the AUTHORS file in the top-level directory for a list of authors.
 *
 * Contact: contact [at] gemfony (dot) eu
 *
 * This file is part of the Geneva library collection.
 *
 * Geneva was developed with kind support from Karlsruhe Institute of
 * Technology (KIT) and Steinbuch Centre for Computing (SCC). Further
 * information about KIT and SCC can be found at http://www.kit.edu/english
 * and http://scc.kit.edu .
 *
 * Geneva is free software: you can redistribute and/or modify it under
 * the terms of version 3 of the GNU Affero General Public License
 * as published by the Free Software Foundation.
 *
 * Geneva is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with the Geneva library. If not, see <http://www.gnu.org/licenses/>.
 *
 * For further information on Gemfony scientific and Geneva, visit
 * http://www.gemfony.eu .
 */


// Standard header files go here

// Boost header files go here
#include <boost/mpl/assert.hpp>

#ifndef GMULTICONSTRAINTT_HPP_
#define GMULTICONSTRAINTT_HPP_

// For Microsoft-compatible compilers
#if defined(_MSC_VER)  &&  (_MSC_VER >= 1020)
#pragma once
#endif

// Geneva header files go here
#include "common/GPODExpectationChecksT.hpp"
#include "geneva/GObject.hpp"
#include "geneva/GOptimizationEnums.hpp"
#include "geneva/GenevaHelperFunctionsT.hpp"

namespace Gem {
namespace Geneva {

// Forward declaration
class GOptimizableEntity;

/******************************************************************************/
/**
 * This is the base class of a hierarchy of classes dealing with inter-parameter
 * constraints. Objects representing the template parameter are evaluated for their
 * validity.
 */
template <typename ind_type>
class GValidityCheckT : public GObject
{
   ///////////////////////////////////////////////////////////////////////
   friend class boost::serialization::access;

   template<typename Archive>
   void serialize(Archive & ar, const unsigned int){
     using boost::serialization::make_nvp;
     ar
     & BOOST_SERIALIZATION_BASE_OBJECT_NVP(GObject);
   }
   ///////////////////////////////////////////////////////////////////////

   // We only accept validity checks for types derived directly or indirectly from GOptimizableEntity
   BOOST_MPL_ASSERT((boost::is_base_of<Gem::Geneva::GOptimizableEntity , ind_type>));

public:
   /***************************************************************************/
   /**
    * The default constructor
    */
   GValidityCheckT()
   { /* nothing */ }

   /***************************************************************************/
   /**
    * The copy constructor
    */
   GValidityCheckT(const GValidityCheckT<ind_type>& cp)
      : GObject(cp)
   { /* nothing */ }

   /***************************************************************************/
   /**
    * The destructor
    */
   virtual ~GValidityCheckT()
   { /* nothing */ }

   /***************************************************************************/
   /**
    * A standard assignment operator
    */
   const GValidityCheckT<ind_type>& operator=(const GValidityCheckT<ind_type>& cp)  {
      GValidityCheckT<ind_type>::load_(&cp);
      return *this;
   }

   /***************************************************************************/
   /**
    * Checks for equality with another GIndividualConstraint object
    */
   bool operator==(const GValidityCheckT<ind_type>& cp) const {
      using namespace Gem::Common;
      // Means: The expectation of equality was fulfilled, if no error text was emitted (which converts to "true")
      return !checkRelationshipWith(cp, CE_EQUALITY, 0.,"GValidityCheckT<ind_type>::operator==","cp", CE_SILENT);
   }

   /***************************************************************************/
   /**
    * Checks for inequality with another GIndividualConstraint object
    */
   bool operator!=(const GValidityCheckT<ind_type>& cp) const {
      using namespace Gem::Common;
      // Means: The expectation of inequality was fulfilled, if no error text was emitted (which converts to "true")
      return !checkRelationshipWith(cp, CE_INEQUALITY, 0.,"GValidityCheckT<ind_type>::operator!=","cp", CE_SILENT);
   }

   /***************************************************************************/
   /**
    * Checks whether a given expectation for the relationship between this object and another object is fulfilled.
    *
    * @param cp A constant reference to another object, camouflaged as a GObject
    * @param e The expected outcome of the comparison
    * @param limit The maximum deviation for floating point values (important for similarity checks)
    * @param caller An identifier for the calling entity
    * @param y_name An identifier for the object that should be compared to this one
    * @param withMessages Whether or not information should be emitted in case of deviations from the expected outcome
    * @return A boost::optional<std::string> object that holds a descriptive string if expectations were not met
    */
   virtual boost::optional<std::string> checkRelationshipWith(
      const GObject& cp
      , const Gem::Common::expectation& e
      , const double& limit
      , const std::string& caller
      , const std::string& y_name
      , const bool& withMessages
   ) const OVERRIDE {
      using namespace Gem::Common;

      // Check that we are indeed dealing with an object of the same type and that we are not
      // accidently trying to compare this object with itself.
      const GValidityCheckT<ind_type> *p_load = GObject::gobject_conversion<GValidityCheckT<ind_type> >(&cp);

      // Will hold possible deviations from the expectation, including explanations
      std::vector<boost::optional<std::string> > deviations;

      // Check our parent class'es data ...
      deviations.push_back(GObject::checkRelationshipWith(cp, e, limit, "GValidityCheckT<ind_type>", y_name, withMessages));

      // no local data

      return evaluateDiscrepancies("GValidityCheckT<ind_type>", caller, deviations, e);
   }

   /***************************************************************************/
   /**
    * Adds local configuration options to a GParserBuilder object
    */
   virtual void addConfigurationOptions(
      Gem::Common::GParserBuilder& gpb
      , const bool& showOrigin
   ) OVERRIDE {
      // Call our parent class'es function
      GObject::addConfigurationOptions(gpb, showOrigin);
   }

   /***************************************************************************/
   /**
    * Checks whether a given parameter set is valid. The function returns a
    * double value which is expected to be larger than 0. Values in the range
    * [0,1] indicate valid parameters (according to this constraint). Values
    * above 1 indicate invalid parameters. The size of the return value can thus
    * be used to indicate the extent of the invalidity.
    */
   double check(
      const ind_type *cp
   ) const {
      double result = check_(cp);

#ifdef DEBUG
      if(result < 0.) {
         glogger
         << "In GValidityCheckT<>::check(): Error!" << std::endl
         << "Validity check yielded a result < 0: " << result << std::endl
         << GEXCEPTION;
      }
#endif /* DEBUG */

      return result;
   }

   /***************************************************************************/
   /**
    * Checks whether the constraint is valid
    *
    * @param cp A pointer to the individual to be checked
    * @param validityLevel Will be filled with the validity level of this individual
    */
   bool isValid(const ind_type *cp, double &validityLevel) const {
      validityLevel = this->check(cp);
      return (validityLevel <= 1.);
   }

protected:
   /***************************************************************************/
   /**
    * Checks whether a given parameter set is valid. The function returns a
    * double value which is expected to be >= 0., giving a level of confidence
    * that this is a valid solution. This function must be overloaded in
    * derived classes.
    */
   virtual double check_(const ind_type *) const = 0;

   /***************************************************************************/
   /**
    * Loads the data of another GValidityCheckT<ind_type>
    */
   virtual void load_(const GObject* cp) OVERRIDE {
      // Check that we are indeed dealing with an object of the same type and that we are not
      // accidently trying to compare this object with itself.
      const GValidityCheckT<ind_type> *p_load = GObject::gobject_conversion<GValidityCheckT<ind_type> >(cp);

      // Load our parent class'es data ...
      GObject::load_(cp);

      // no local data
   }

   /***************************************************************************/
   /** @brief Creates a deep clone of this object */
   virtual GObject* clone_() const = 0;

   /***************************************************************************/
};

/******************************************************************************/
////////////////////////////////////////////////////////////////////////////////
/******************************************************************************/
/**
 * An collection of validity checks with the GValidityCheckT interface
 */
template <typename ind_type>
class GValidityCheckContainerT : public GValidityCheckT<ind_type>
{
   ///////////////////////////////////////////////////////////////////////
   friend class boost::serialization::access;

   template<typename Archive>
   void serialize(Archive & ar, const unsigned int){
     using boost::serialization::make_nvp;
     ar
     & BOOST_SERIALIZATION_BASE_OBJECT_NVP(GValidityCheckT<ind_type>);
   }
   ///////////////////////////////////////////////////////////////////////

public:
   /***************************************************************************/
   /**
    * The default constructor
    */
   GValidityCheckContainerT()
   { /* nothing */ }

   /***************************************************************************/
   /**
    * Initialization from a vector of validity checks
    */
   GValidityCheckContainerT(const std::vector<boost::shared_ptr<GValidityCheckT<ind_type> > >& validityChecks)
   {
      copyGenevaSmartPointerVector(validityChecks, validityChecks_);
   }

   /***************************************************************************/
   /**
    * The copy constructor
    */
   GValidityCheckContainerT(const GValidityCheckContainerT<ind_type>& cp)
      : GValidityCheckT<ind_type>(cp)
   {
      copyGenevaSmartPointerVector(cp.validityChecks_, validityChecks_);
   }

   /***************************************************************************/
   /**
    * The destructor
    */
   virtual ~GValidityCheckContainerT()
   { /* nothing */ }

   /***************************************************************************/
   /**
    * A standard assignment operator
    */
   const GValidityCheckContainerT<ind_type>& operator=(const GValidityCheckContainerT<ind_type>& cp)  {
      GValidityCheckContainerT<ind_type>::load_(&cp);
      return *this;
   }

   /***************************************************************************/
   /**
    * Checks for equality with another GValidityCheckContainerT object
    */
   bool operator==(const GValidityCheckContainerT<ind_type>& cp) const {
      using namespace Gem::Common;
      // Means: The expectation of equality was fulfilled, if no error text was emitted (which converts to "true")
      return !checkRelationshipWith(cp, CE_EQUALITY, 0.,"GValidityCheckContainerT<ind_type>::operator==","cp", CE_SILENT);
   }

   /***************************************************************************/
   /**
    * Checks for inequality with another GValidityCheckContainerT object
    */
   bool operator!=(const GValidityCheckContainerT<ind_type>& cp) const {
      using namespace Gem::Common;
      // Means: The expectation of inequality was fulfilled, if no error text was emitted (which converts to "true")
      return !checkRelationshipWith(cp, CE_INEQUALITY, 0.,"GValidityCheckContainerT<ind_type>::operator!=","cp", CE_SILENT);
   }

   /***************************************************************************/
   /**
    * Checks whether a given expectation for the relationship between this object and another object is fulfilled.
    *
    * @param cp A constant reference to another object, camouflaged as a GObject
    * @param e The expected outcome of the comparison
    * @param limit The maximum deviation for floating point values (important for similarity checks)
    * @param caller An identifier for the calling entity
    * @param y_name An identifier for the object that should be compared to this one
    * @param withMessages Whether or not information should be emitted in case of deviations from the expected outcome
    * @return A boost::optional<std::string> object that holds a descriptive string if expectations were not met
    */
   virtual boost::optional<std::string> checkRelationshipWith(
      const GObject& cp
      , const Gem::Common::expectation& e
      , const double& limit
      , const std::string& caller
      , const std::string& y_name
      , const bool& withMessages
   ) const OVERRIDE {
      using namespace Gem::Common;

      // Check that we are indeed dealing with an object of the same type and that we are not
      // accidently trying to compare this object with itself.
      const GValidityCheckContainerT<ind_type> *p_load = GObject::gobject_conversion<GValidityCheckContainerT<ind_type> >(&cp);

      // Will hold possible deviations from the expectation, including explanations
      std::vector<boost::optional<std::string> > deviations;

      // Check our parent class'es data ...
      deviations.push_back(GValidityCheckT<ind_type>::checkRelationshipWith(cp, e, limit, "GValidityCheckContainerT<ind_type>", y_name, withMessages));

      // ... and then our local data
      deviations.push_back(checkExpectation(withMessages, "GValidityCheckContainerT<ind_type>", validityChecks_, p_load->validityChecks_, "validityChecks_", "p_load->validityChecks_", e , limit));

      return evaluateDiscrepancies("GValidityCheckContainerT<ind_type>", caller, deviations, e);
   }

   /***************************************************************************/
   /**
    * Adds a validity check to this object. Note that we clone the check so
    * that it can be used multiple times.
    */
   void addCheck(boost::shared_ptr<GValidityCheckT<ind_type> > vc_ptr) {
      if(!vc_ptr) {
         glogger
         << "In GValidityCheckContainerT<>::addCheck(): Error!" << std::endl
         << "Got empty check pointer" << std::endl
         << GEXCEPTION;
      }

      validityChecks_.push_back(vc_ptr->GObject::template clone<GValidityCheckT<ind_type> >());
   }

protected:
   /***************************************************************************/
   /** @brief Checks whether a given parameter set is valid. To be specified in derived classes */
   virtual double check_(const ind_type *) const = 0;

   /***************************************************************************/
   /** @brief Creates a deep clone of this object */
   virtual GObject* clone_() const = 0;

   /***************************************************************************/
   /**
    * Loads the data of another GValidityCheckT<ind_type>
    */
   virtual void load_(const GObject* cp) OVERRIDE {
      // Check that we are indeed dealing with an object of the same type and that we are not
      // accidently trying to compare this object with itself.
      const GValidityCheckContainerT<ind_type> *p_load = GObject::gobject_conversion<GValidityCheckContainerT<ind_type> >(cp);

      // Load our parent class'es data ...
      GValidityCheckT<ind_type>::load_(cp);

      // and then our local data
      copyGenevaSmartPointerVector(p_load->validityChecks_, validityChecks_);
   }

   /***************************************************************************/
   /** @brief Holds all registered validity checks */
   std::vector<boost::shared_ptr<GValidityCheckT<ind_type> > > validityChecks_;
};

/******************************************************************************/
////////////////////////////////////////////////////////////////////////////////
/******************************************************************************/
/**
 * A class which combines all values (i.e. values > 1) according to a
 * user-defined policy or returns 0, if all checks are valid.
 */
template <typename ind_type>
class GCheckCombinerT
   : public GValidityCheckContainerT<ind_type>
{
   ///////////////////////////////////////////////////////////////////////
   friend class boost::serialization::access;

   template<typename Archive>
   void serialize(Archive & ar, const unsigned int){
     using boost::serialization::make_nvp;
     ar
     & BOOST_SERIALIZATION_BASE_OBJECT_NVP(GValidityCheckT<ind_type>)
     & BOOST_SERIALIZATION_NVP(combinerPolicy_);
   }
   ///////////////////////////////////////////////////////////////////////

public:
   /***************************************************************************/
   /**
    * The default constructor
    */
   GCheckCombinerT()
      : combinerPolicy_(Gem::Geneva::MULTIPLYINVALID)
   { /* nothing */ }

   /***************************************************************************/
   /**
    * Initialization from a vector of validity checks
    */
   GCheckCombinerT(
      const std::vector<boost::shared_ptr<GValidityCheckT<ind_type> > >& validityChecks
   )
      : GValidityCheckContainerT<ind_type>(validityChecks)
      , combinerPolicy_(Gem::Geneva::MULTIPLYINVALID)
   { /* nothing */ }

   /***************************************************************************/
   /**
    * The copy constructor
    */
   GCheckCombinerT(const GCheckCombinerT<ind_type>& cp)
      : GValidityCheckContainerT<ind_type>(cp)
      , combinerPolicy_(cp.combinerPolicy_)
   { /* nothing */ }

   /***************************************************************************/
   /**
    * The destructor
    */
   virtual ~GCheckCombinerT()
   { /* nothing */ }

   /***************************************************************************/
   /**
    * A standard assignment operator
    */
   const GCheckCombinerT<ind_type>& operator=(const GCheckCombinerT<ind_type>& cp)  {
      GCheckCombinerT<ind_type>::load_(&cp);
      return *this;
   }

   /***************************************************************************/
   /**
    * Checks for equality with another GCheckCombinerT object
    */
   bool operator==(const GCheckCombinerT<ind_type>& cp) const {
      using namespace Gem::Common;
      // Means: The expectation of equality was fulfilled, if no error text was emitted (which converts to "true")
      return !checkRelationshipWith(cp, CE_EQUALITY, 0.,"GCheckCombinerT<ind_type>::operator==","cp", CE_SILENT);
   }

   /***************************************************************************/
   /**
    * Checks for inequality with another GCheckCombinerT object
    */
   bool operator!=(const GCheckCombinerT<ind_type>& cp) const {
      using namespace Gem::Common;
      // Means: The expectation of inequality was fulfilled, if no error text was emitted (which converts to "true")
      return !checkRelationshipWith(cp, CE_INEQUALITY, 0.,"GCheckCombinerT<ind_type>::operator!=","cp", CE_SILENT);
   }

   /***************************************************************************/
   /**
    * Checks whether a given expectation for the relationship between this object and another object is fulfilled.
    *
    * @param cp A constant reference to another object, camouflaged as a GObject
    * @param e The expected outcome of the comparison
    * @param limit The maximum deviation for floating point values (important for similarity checks)
    * @param caller An identifier for the calling entity
    * @param y_name An identifier for the object that should be compared to this one
    * @param withMessages Whether or not information should be emitted in case of deviations from the expected outcome
    * @return A boost::optional<std::string> object that holds a descriptive string if expectations were not met
    */
   virtual boost::optional<std::string> checkRelationshipWith(
      const GObject& cp
      , const Gem::Common::expectation& e
      , const double& limit
      , const std::string& caller
      , const std::string& y_name
      , const bool& withMessages
   ) const OVERRIDE {
      using namespace Gem::Common;

      // Check that we are indeed dealing with an object of the same type and that we are not
      // accidently trying to compare this object with itself.
      const GCheckCombinerT<ind_type> *p_load = GObject::gobject_conversion<GCheckCombinerT<ind_type> >(&cp);

      // Will hold possible deviations from the expectation, including explanations
      std::vector<boost::optional<std::string> > deviations;

      // Check our parent class'es data ...
      deviations.push_back(GValidityCheckContainerT<ind_type>::checkRelationshipWith(cp, e, limit, "GCheckCombinerT<ind_type>", y_name, withMessages));

      // ... and then our local data
      deviations.push_back(checkExpectation(withMessages, "GCheckCombinerT<ind_type>", combinerPolicy_, p_load->combinerPolicy_, "combinerPolicy_", "p_load->combinerPolicy_", e , limit));

      return evaluateDiscrepancies("GCheckCombinerT<ind_type>", caller, deviations, e);
   }

   /***************************************************************************/
   /**
    * Allows to set the combiner policy
    */
   void setCombinerPolicy(validityCheckCombinerPolicy combinerPolicy) {
      combinerPolicy_ = combinerPolicy;
   }

   /***************************************************************************/
   /**
    * Allows to retrieve the combiner policy
    */
   validityCheckCombinerPolicy getCombinerPolicy() const {
      return combinerPolicy_;
   }

protected:
   /***************************************************************************/
   /**
    * Combines all parameters according to a user-defined policy
    */
   virtual double check_(const ind_type *cp) const {
      switch(combinerPolicy_) {
         // --------------------------------------------------------------------
         case Gem::Geneva::MULTIPLYINVALID:
         {
            std::vector<double> invalidChecks;
            double validityLevel;
            typename std::vector<boost::shared_ptr<GValidityCheckT<ind_type> > >::const_iterator cit;
            for(cit=GValidityCheckContainerT<ind_type>::validityChecks_.begin(); cit!=GValidityCheckContainerT<ind_type>::validityChecks_.end(); ++cit) {
               if(!(*cit)->isValid(cp, validityLevel)) {
                  invalidChecks.push_back(validityLevel);
               }
            }
            if(invalidChecks.empty()) { // All checks were valid
               return 0.;
            } else { // There were invalid results
               double result = 1.;
               std::vector<double>::const_iterator d_cit;
               for(d_cit=invalidChecks.begin(); d_cit!=invalidChecks.end(); ++d_cit) {
#ifdef DEBUG
                  if(1. >= *d_cit) {
                     glogger
                     << "In GCheckCombinerT<>::check_(): Error!" << std::endl
                     << "Result " << *d_cit << " is marked as invalid although it is <= 1." << std::endl
                     << GEXCEPTION;
                  }
#endif

                  result *= *d_cit;
               }
               return result;
            }
         }
            break;

         // --------------------------------------------------------------------
         default:
         {
            glogger
            << "In GCheckCombinerT<ind_type>::check_(): Error!" << std::endl
            << "Got invalid combinerPolicy_ value: " << combinerPolicy_ << std::endl
            << GEXCEPTION;
         }
            break;
         // --------------------------------------------------------------------
      }

      // Make the compiler happy
      return 0.;
   }

   /***************************************************************************/
   /**
    * Creates a deep clone of this object
    */
   virtual GObject* clone_() const {
      return new GCheckCombinerT<ind_type>(*this);
   }

   /***************************************************************************/
   /**
    * Loads the data of another GValidityCheckT<ind_type>
    */
   virtual void load_(const GObject* cp) OVERRIDE {
      // Check that we are indeed dealing with an object of the same type and that we are not
      // accidently trying to compare this object with itself.
      const GCheckCombinerT<ind_type> *p_load = GObject::gobject_conversion<GCheckCombinerT<ind_type> >(cp);

      // Load our parent class'es data ...
      GValidityCheckT<ind_type>::load_(cp);

      // and then our local data
      combinerPolicy_ = p_load->combinerPolicy_;
   }

private:
   /***************************************************************************/
   // Local data

   validityCheckCombinerPolicy combinerPolicy_; ///< Indicates how validity checks should be combined
};

/******************************************************************************/
////////////////////////////////////////////////////////////////////////////////
/******************************************************************************/

} /* namespace Geneva */
} /* namespace Gem */

/******************************************************************************/
// The content of BOOST_SERIALIZATION_ASSUME_ABSTRACT(T)
namespace boost {
   namespace serialization {
      template<typename ind_type>
      struct is_abstract< Gem::Geneva::GValidityCheckT<ind_type> > : public boost::true_type {};
      template<typename ind_type>
      struct is_abstract< const Gem::Geneva::GValidityCheckT<ind_type> > : public boost::true_type {};
   }
}

namespace boost {
   namespace serialization {
      template<typename ind_type>
      struct is_abstract< Gem::Geneva::GValidityCheckContainerT<ind_type> > : public boost::true_type {};
      template<typename ind_type>
      struct is_abstract< const Gem::Geneva::GValidityCheckContainerT<ind_type> > : public boost::true_type {};
   }
}

/******************************************************************************/

#endif /* GMULTICONSTRAINTT_HPP_ */
